# Function that provides the iterative integrals

def iter_int(u,t0, tf, dt, Ntrunc):
    import numpy as np
    
    """"
    Returns the list of all iterated integrals of the input u indexed by the words of length from 1 to Ntrunc.
    
    
    Parameters:
    -----------
    u: array_like
        The array of input functions u_i: [t0, tf] -> IR, for all i in {1, ..., m}
        stacked vertically. Each function has the form u_i = np.array([u_i[0], u_i[1], ..., u_i[N]])  
        where u_i[0] = u_i(t_0), u_i[1] = u_i(t_0+dt), ..., u_i[N] = u_i(tf), N = int((tf-t0)//dt+1)
        and u = np.vstack([u_1, u_2, ..., u_m])
    
    t0: float
        Initial point of the time-interval domain of the inputs
    
    tf: float
        Final time of the time-interval domain of the inputs
    
    dt: float
        The size of the step of the evenly spaced partition of the time-interval domain 
    
    Ntrunc: int
        The truncation length of the words of the Chen-Fliess series
        Sum_{i=0}^Ntrunc Sum_{eta in X^i} (c, eta) E_{eta}[u](t0,tf)
    
    
    
    Returns:
    --------
    list: ndarray
    
    [
    [E_{x_0}[u](t0, t0), E_{x_0}[u](t0, t0+dt), ..., E_{x_0}[u](t0, tf)],
    [E_{x_1}[u](t0, t0), E_{x_1}[u](t0, t0+dt), ..., E_{x_1}[u](t0, tf)],
     .
     .
     .
    [E_{x_m}[u](t0, t0), E_{x_m}[u](t0, t0+dt), ..., E_{x_m}[u](t0, tf)],
    ---------------------------------------------------------------------------------------
    [E_{x_0x_0}[u](t0, t0), E_{x_0x_0}[u](t0, t0+dt), ..., E_{x_0x_0}[u](t0, tf)],
    [E_{x_1x_0}[u](t0, t0), E_{x_1x_0}[u](t0, t0+dt), ..., E_{x_1x_0}[u](t0, tf)],
     .
     .
     .
    [E_{x_mx_0}[u](t0, t0), E_{x_mx_0}[u](t0, t0+dt), ..., E_{x_mx_0}[u](t0, tf)],
    [E_{x_0x_1}[u](t0, t0), E_{x_0x_1}[u](t0, t0+dt), ..., E_{x_0x_1}[u](t0, tf)],
    [E_{x_1x_1}[u](t0, t0), E_{x_1x_1}[u](t0, t0+dt), ..., E_{x_1x_1}[u](t0, tf)],
     .
     .
     .
    [E_{x_mx_1}[u](t0, t0), E_{x_mx_1}[u](t0, t0+dt), ..., E_{x_mx_1}[u](t0, tf)],
     .
     .
     .
    [E_{x_0x_m}[u](t0, t0), E_{x_0x_m}[u](t0, t0+dt), ..., E_{x_0x_m}[u](t0, tf)],
    [E_{x_1x_m}[u](t0, t0), E_{x_1x_m}[u](t0, t0+dt), ..., E_{x_1x_m}[u](t0, tf)],
     .
     .
     .
    [E_{x_mx_m}[u](t0, t0), E_{x_mx_m}[u](t0, t0+dt), ..., E_{x_mx_m}[u](t0, tf)],
    ----------------------------------------------------------------------------------------
    [E_{x_0x_0x_0}[u](t0, t0), E_{x_0x_0x_0}[u](t0, t0+dt), ..., E_{x_0x_0x_0}[u](t0, tf)],
    [E_{x_0x_1x_0}[u](t0, t0), E_{x_0x_1x_0}[u](t0, t0+dt), ..., E_{x_0x_1x_0}[u](t0, tf)],
     .
     .
     .
    [E_{x_0x_mx_0}[u](t0, t0), E_{x_0x_mx_0}[u](t0, t0+dt), ..., E_{x_0x_mx_0}[u](t0, tf)],
     .
     .
     .
    [E_{x_mx_mx_m}[u](t0, t0), E_{x_mx_mx_m}[u](t0, t0+dt), ..., E_{x_mx_mx_m}[u](t0, tf)]
    ----------------------------------------------------------------------------------------
     .
     .
     .
    ----------------------------------------------------------------------------------------
    [E_{x_0...x_0}[u](t0, t0), E_{x_0...x_0}[u](t0, t0+dt), ..., E_{x_0...x_0}[u](t0, tf)],
     .
     .
     .
    [E_{x_m...x_m}[u](t0, t0), E_{x_m...x_m}[u](t0, t0+dt), ..., E_{x_m...x_m}[u](t0, tf)],
    ]
    
    
    
    Examples:
    ---------
    
    import numpy as np
    from scipy.integrate import solve_ivp
    import matplotlib.pyplot as plt

    # Define the system
    def system(t, x, u1_func, u2_func):
        x1, x2 = x
        u1 = u1_func(t)
        u2 = u2_func(t)
        dx1 = -x1*x2 +  x1 * u1
        dx2 = x1*x2 - x2* u2
        return [dx1, dx2]

    # Input 1
    def u1_func(t):
        return np.sin(t)

    # Input 2
    def u2_func(t):
        return np.cos(t)

    # Initial condition
    x0 = [1/6,1/6]

    # Time range
    t_span = (0, 3)

    # Solve the ODE with dense_output=True
    solution = solve_ivp(system, t_span, x0, args=(u1_func, u2_func), dense_output=True)

    # Partition of the time range
    t = np.linspace(t_span[0], t_span[1], 300)
    y = solution.sol(t)

    # Plot the results
    plt.plot(y[0].T, y[1].T)
    plt.xlabel('$x_1$')
    plt.ylabel('$x_2$')
    plt.grid()
    plt.show()

    plt.plot(t, y[0].T)
    plt.xlabel('Time')
    plt.ylabel('$x_1$')
    plt.grid()
    plt.show()
    
    """
    
    
    if t0 < 0:
        raise ValueError("The initial point, %s, must be non-negative." %t0)
        
    if tf < 0:
        raise ValueError("The final point, %s, must be non-negative." %tf)
        
    if dt < 0:
        raise ValueError("The step, %s, must be non-negative." %dt)
        
    if Ntrunc < 1:
        raise ValueError("The truncation length, %s, must be non-negative." %Ntrunc)
        
    if not isinstance(Ntrunc, int):
        raise ValueError("The truncation length, %s, must be an integer." %Ntrunc)
        
    if not isinstance(u, np.ndarray):
        raise ValueError("The input, %s, must be an ndarray." %u)
        
    if u.shape[0] == 0:
        raise ValueError("The number of rows, %s, must be greater than zero." %u.shape[0])
        

        
        
    # The length of the partition of time is computed
    length_t = int((tf-t0)//dt+1)
    
    if u.shape[1] != length_t:
        raise ValueError("The length of the input, %s, must be int((tf-t0)//dt+1) = %s." %(u.shape[1], length_t))
    
    # The sequence of numbers that represent the partition of the time domain of the input is generated.
    # This is used to discretize the integrals.
    t = np.linspace(t0, tf, length_t)
    # The input u_0 associated with the letter x_0 is generated.
    u0 = np.ones(length_t)
    # The inputs of the system associated with x_1, ..., x_m with the input associated to x_0 vertically are stacked.
    # [
    #  [u_0(t0), u_0(t0+dt), ..., u_0(tf)],
    #  [u_1(t0), u_1(t0+dt), ..., u_1(tf)],
    #    .
    #    .
    #    .
    #  [u_m(t0), u_m(t0+dt), ..., u_m(tf)]
    # ]
    u = np.vstack([u0, u])
    
    # The number of rows which are equal to the number of total input functions is obtained.
    num_input = int(np.size(u,0))

    # The total number of iterated integrals of word length less than or equal to the truncation length is computed.
    # total_iterint = num_input + num_input**2 + ... + num_input**Ntrunc
    total_iterint = num_input*(1-pow(num_input,Ntrunc))/(1-num_input)
    # This is transformed into an integer.
    total_iterint = int(total_iterint)
    
    # A matrix of zeros with as many rows as the total number of iterated integrals and as many columns
    # as the elements in the partition of time is computed.
    Etemp = np.zeros((total_iterint,length_t))
    # Starts the list ctrEtemp such that ctrEtemp[k]-ctrEtemp[k-1] = the number of iterated integrals 
    # of word length k
    # ctrEtemp[0] = 0
    # ctrEtemp[k] = num_input + num_input**2 + ... + num_input**k,  1<=k<=Ntrunc
    ctrEtemp = np.zeros(Ntrunc+1)

    # ctrEtemp[k] = num_input + num_input**2 + ... + num_input**k,  1<=k<=Ntrunc
    for i in range(Ntrunc):
        ctrEtemp[i+1] = ctrEtemp[i]+pow(num_input,i+1)
    
    
    # The iterative integrals of the words of length 1, E_{x_i}[u](t0, tf) for all i in {0, ..., m}, are computed.
    
    # First, E_{x_i}[u](t0, tf) for all i in {0, ..., m} are computed 
    # for all tf neq t0
    sum_acc = np.cumsum(u, axis = 1)*dt
    # Then the values of E_{x_i}[u](t0, tf) for tf = t0, this is E_{x_i}[u](t0, t0) = 0, are added
    # to have E_{x_i}[u](t0, tf) for all tf>=t0
    Etemp[:num_input,:] = np.hstack((np.zeros((num_input,1)), sum_acc[:,:-1]))
    
    
    # The iterated integrals of the words of length k => 1, E_{x_{i_1}...x_{i_k}}[u](t0, tf) for all i_j in {0, ..., m}, 
    # are computed at each iteration.
    
    for i in range(1,Ntrunc):
        # start_prev_block = num_input + num_input**2 + ... + num_input**(i-1)
        start_prev_block = int(ctrEtemp[i-1])
        # end_prev_block = num_input + num_input**2 + ... + num_input**i
        end_prev_block = int(ctrEtemp[i])
        # end_current_block = num_input + num_input**2 + ... + num_input**(i+1)
        end_current_block = int(ctrEtemp[i+1])
        # num_prev_block = num_input**i
        num_prev_block = end_prev_block - start_prev_block
        # num_current_block = num_input**(i+1)
        num_current_block = end_current_block - end_prev_block
        
        """ 
        U_block =
        [
         u_0
         u_0
          .   
          .   # u_0 repeats num_input**i times
          . 
         u_0
         u_1
         u_1
          .
          .   # u_1 repeats num_input**i times
          .
         u_1
          .
          .   # u_k repeats num_input**i times
          .
         u_m
         u_m
          .
          .   # u_m repeats num_input**i times
          .
         u_m
        ]
        """
        U_block = u[np.repeat(range(num_input), num_prev_block), :]
        
        """
        prev_int_block =
        [
        E_{x_0...x_0x_0}[u](t0,tf)  
        E_{x_0...x_1x_0}[u](t0,tf)
                .
                .                         # block with all the num_input**i iterated integrals of words of length i
                .
        E_{x_m...x_mx_m}[u](t0,tf)
        --------------------------
        E_{x_0...x_0x_0}[u](t0,tf)  
        E_{x_0...x_1x_0}[u](t0,tf)
                .
                .                         # block with all the num_input**i iterated integrals of words of length i
                .
        E_{x_m...x_mx_m}[u](t0,tf)
        --------------------------
                .
                .
                .
        --------------------------
        E_{x_0...x_0x_0}[u](t0,tf)  
        E_{x_0...x_1x_0}[u](t0,tf)
                .
                .                         # block with all the num_input**i iterated integrals of words of length i
                .
        E_{x_m...x_mx_m}[u](t0,tf)
        ]
        
        In total there are num_input blocks 
        """
        prev_int_block = np.tile(Etemp[start_prev_block:end_prev_block,:],(num_input,1))
        
        
        """
        U_block*prev_int_block = 
        [
        u_0(tf)E_{x_0...x_0x_0}[u](t0,tf)  
        u_0(tf)E_{x_0...x_1x_0}[u](t0,tf)
                .
                .                         # block with all the num_input**i iterated integrals of words of length i
                .
        u_0(tf)E_{x_m...x_mx_m}[u](t0,tf)
        --------------------------
        u_1(tf)E_{x_0...x_0x_0}[u](t0,tf)  
        u_1(tf)E_{x_0...x_1x_0}[u](t0,tf)
                .
                .                         # block with all the num_input**i iterated integrals of words of length i
                .
        u_1(tf)E_{x_m...x_mx_m}[u](t0,tf)
        --------------------------
                .
                .
                .
        --------------------------
        u_m(tf)E_{x_0...x_0x_0}[u](t0,tf)  
        u_m(tf)E_{x_0...x_1x_0}[u](t0,tf)
                .
                .                         # block with all the num_input**i iterated integrals of words of length i
                .
        u_m(tf)E_{x_m...x_mx_m}[u](t0,tf)
        ]
        
        current_int_block integrates U_block*prev_int_block
        """
        current_int_block = np.cumsum(U_block*prev_int_block, axis = 1)*dt
        # Stacks the block of iterated integrals of word length i+1 into Etemp
        Etemp[end_prev_block:end_current_block,:] = np.hstack((np.zeros((num_current_block,1)), current_int_block[:,:-1]))

    itint = Etemp
    return itint                                                                        





def single_iter_lie(eta, h,vector_field,z):
    import numpy as np
    import sympy as sp

    # The list that will contain all the Lie derivatives is initiated. 
    LT = sp.Matrix([h])
    
    for i in eta:
        
        LT = LT.jacobian(z)*vector_field[:,i]
    return LT



# Function that provides the iterative Lie derivatives
def iter_lie(h,vector_field,z,Ntrunc):
    import numpy as np
    import sympy as sp

    """
    Returns the list of all the Lie derivatives indexed by the words of length from 1 to Ntrunc
    Given the system 
        dot{z} = g_0(z) + sum_{i=1}^m g_i(z) u_i(t), 
        y = h(z)
    with g_i: S -> IR^n and h: S -> IR, S is a subset of IR^n for all i in {0, ..., n}
    The Lie derivative L_eta h of the output function h(z) indexed by the word 
    eta = x_{i_1}x_{i_2}...x_{i_k} is defined recursively as
        L_eta h = L_(x_{i_2}...x_{i_k}) (partial/partial z h) cdot g_{i_1} 
    
    
    
    Parameters:
    -----------
    h: symbolic
        The symbolic function that represents the output of the system.
    
    vector_field: symbolic array
        The array that contains the vector fields of the system.
        vector_field = 
            sp.transpose(sp.Matrix([[(g_0)_1, ..., (g_0)_n],[(g_1)_1, ..., (g_1)_n], ..., [(g_m)_1, ..., (g_m)_n]]))
    
    z: symbolic array
        The domain of the vector fields.
        z = sp.Matrix([z1, z2, ...., zn])
    
    Ntrunc: int
        The truncation length of the words index of the Lie derivatives
    
    
    
    Returns:
    --------
    list: symbolic array
    
    [
    L_{x_0} h
     .
     .
     .
    L_{x_m} h
    ---------------
    L_{x_0x_0} h
     .
     .
     .
    L_{x_mx_m} h
    ---------------
     .
     .
     .
    ---------------
    L_{x_0...x_0} h
     .
     .
     .
    L_{x_m...x_m} h
    ]
    
    
    
    Examples:
    ---------
    
    # Consider the system:
    # dx1 = -x1*x2 +  x1 * u1
    # dx2 = x1*x2 - x2* u2
    
    # Define the state of the system:
    x1, x2 = sp.symbols('x1 x2')
    x = sp.Matrix([x1, x2])

    # h is a real value function representing the output.
    h = x1

    # each column of g = [[g10, g11, g12, ..., g1m], [g20, g21, g22, ..., g2m], ..., [gn1, gn2, gn3, ..., gnm]]
    # is a vector field where xdot = g0 + g_1*u_1 + g_2*u_2 +...+ g_m*u_m
    # g0 = transpose([g10, g20, ...., gn0])
    # g_1 = transpose([g11, g21, ...., gn1])
    # g_2 = transpose([g12, g22, ...., gn2])
    #                .
    #                .
    #                .
    # g_m = transpose([g0m, g1m, ...., gnm])

    # Define the vector field
    g = sp.transpose(sp.Matrix([[-x1*x2, x1*x2], [x1, 0], [0, - x2]]))

    # Ntrunc is the maximum length of words
    Ntrunc = 4

    Ceta = np.array(iter_lie(h,g,x,Ntrunc).subs([(x[0], 1/6),(x[1], 1/6)]))
    Ceta
    
    
    """
    
    if h is None:
        raise ValueError("The output function, %s , cannot be null" %h)
        
    if vector_field is None:
        raise ValueError("The vector_field, %s, cannot be null" %vector_field)
        
    if z is None:
        raise ValueError("The argument of output function and the vector field, %s, cannot be null" %z)
    
    if Ntrunc < 1:
        raise ValueError("The truncation length, %s, must be non-negative." %Ntrunc)
    
    if not isinstance(Ntrunc, int):
        raise ValueError("The truncation length, %s, must be an integer." %Ntrunc)
    
    
    # The number of vector fields is obtained.
    # num_vfield = m
    num_vfield = np.size(vector_field,1)
    
    
    # The total number of Lie derivatives of word length less than or equal to the truncation length is computed.
    # total_lderiv = num_input + num_input**2 + ... + num_input**Ntrunc
    total_lderiv = num_vfield*(1-pow(num_vfield, Ntrunc))/(1-num_vfield)
    total_lderiv = int(total_lderiv)
    
    
    # The list that will contain all the Lie derivatives is initiated. 
    Ltemp = sp.Matrix(np.zeros((total_lderiv, 1), dtype='object'))
    ctrLtemp = np.zeros((Ntrunc+1,1), dtype = 'int')
    
    
    # ctrLtemp[k] = num_input + num_input**2 + ... + num_input**k,  1<=k<=Ntrunc
    for i in range(Ntrunc):
        ctrLtemp[i+1] = ctrLtemp[i] + num_vfield**(i+1)
    
    
    # The Lie derivative L_eta h(z) of words eta of length 1 are computed 
    LT = sp.Matrix([h]).jacobian(z)*vector_field

    # Transforms the lie derivative from a row vector to a column vector
    LT = LT.reshape(LT.shape[0]*LT.shape[1], 1)
    
    # Adds the computed Lie derivatives to a repository
    Ltemp[:num_vfield, 0] = LT

    
    # The Lie derivatives of the words of length k => 1, L_{x_{i_1}...x_{i_k}}h(z) for all i_j in {0, ..., m}, 
    # are computed at each iteration.
    
    for i in range(1, Ntrunc):
        # start_prev_block = num_input + num_input**2 + ... + num_input**(i-1)
        start_prev_block = int(ctrLtemp[i-1])
        # end_prev_block = num_input + num_input**2 + ... + num_input**i
        end_prev_block = int(ctrLtemp[i])
        # end_current_block = num_input + num_input**2 + ... + num_input**(i+1)
        end_current_block = int(ctrLtemp[i+1])
        # num_prev_block = num_input**i
        num_prev_block = end_prev_block - start_prev_block
        # num_current_block = num_input**(i+1)
        num_current_block = end_current_block - end_prev_block
    
        """
        LT = 
        [
         [L_{x_0...x_0x_0}h(z)],
         [L_{x_0...x_1x_0}h(z)],
                .
                .
                .
         [L_{x_m...x_m}h(z)],
        ]
        these are the Lie derivatives indexed by words of length i
        """
        LT = Ltemp[start_prev_block:end_prev_block,0]
        
        """
        LT = 
        [
         [partial/ partial z L_{x_0...x_0x_0}h(z)],
         [partial/ partial z L_{x_0...x_1x_0}h(z)],
                .
                .
                .
         [partial/ partial z L_{x_m...x_m}h(z)],
        ]
        * 
        [g_0, g_1, ..., g_m]
        
        =
        
        [
            L_{x_0x_0...x_0}h(z)    |   L_{x_0x_0..x_0x_1x_0}h(z)   |          |  L_{x_0x_m..x_mx_m}h(z)
            L_{x_1x_0...x_0}h(z)    |   L_{x_1x_0..x_0x_1x_0}h(z)   |          |  L_{x_1x_m..x_mx_m}h(z)
            L_{x_2x_0...x_0}h(z)    |   L_{x_2x_0..x_0x_1x_0}h(z)   |          |  L_{x_2x_m..x_mx_m}h(z)
                .                   |        .                      |    ...   |       .
                .                   |        .                      |    ...   |       .
                .                   |        .                      |    ...   |       .
            L_{x_mx_0...x_0}h(z)    |   L_{x_mx_0..x_0x_1x_0}h(z)   |          |  L_{x_mx_m..x_mx_m}h(z)
        ]
        
        """
        LT = LT.jacobian(z)*vector_field
        # Transforms the lie derivative from a row vector to a column vector
        
        """
        LT =
        [
            L_{x_0x_0...x_0}h(z) 
            L_{x_1x_0...x_0}h(z) 
            L_{x_2x_0...x_0}h(z) 
                .                
                .                
                .                
            L_{x_mx_0...x_0}h(z) 
            ------------------------
            L_{x_0x_0...x_1x_0}h(z) 
            L_{x_1x_0...x_1x_0}h(z) 
            L_{x_2x_0...x_1x_0}h(z) 
                .                
                .                
                .                
            L_{x_mx_0...x_1x_0}h(z)
            ------------------------
                .
                .
                .
            ------------------------
            L_{x_0x_m...x_m}h(z) 
            L_{x_1x_m...x_m}h(z) 
            L_{x_2x_m...x_m}h(z) 
                .                
                .                
                .                
            L_{x_mx_m...x_m}h(z) 
        ]
        these are the Lie derivatives indexed by words of length i+1
        """
        LT = LT.reshape(LT.shape[0]*LT.shape[1], 1)
        # Adds the computed Lie derivatives to the repository
        Ltemp[end_prev_block:end_current_block,:]=LT

    return Ltemp





# Individual iterative integral
def single_iter_int(eta, u, t0, tf, dt):
    import numpy as np
    
    # The length of the partition of time is computed
    length_t = int((tf-t0)//dt+1)
    # The sequence of numbers that represent the partition of the time domain of the input is generated.
    # This is used to discretize the integrals.
    t = np.linspace(t0, tf, length_t)
    # The input u_0 associated with the letter x_0 is generated.
    u0 = np.ones(length_t)
    # The inputs of the system associated with x_1, ..., x_m with the input associated to x_0 vertically are stacked.
    # [
    #  [u_0(t0), u_0(t0+dt), ..., u_0(tf)],
    #  [u_1(t0), u_1(t0+dt), ..., u_1(tf)],
    #    .
    #    .
    #    .
    #  [u_m(t0), u_m(t0+dt), ..., u_m(tf)]
    # ]
    u = np.vstack([u0, u])
    temp = 1
    for i in np.flip(eta):
        temp = temp * u[i]
        temp = np.append(0,np.cumsum(temp)*delta)
        temp = temp[:-1]

    return temp




from setuptools import setup, find_packages
import codecs
import os

VERSION = '1.0.12'
DESCRIPTION = 'Chen-Fliess series computation'

with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Setting up
setup(
    name="CFSpy",
    version=VERSION,
    author="Ivan Perez Avellaneda",
    author_email="<iperezave@gmail.com>",
    license="MIT",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/iperezav/CFSpy/',
    packages=find_packages(),
    install_requires=['numpy', 'sympy'],
    keywords=['Chen-Fliess series', 'nonlinear system', 'input-output system', 'ODEs', 'control system', 'system theory', 'python'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)
